import os
import sys
import numpy as np
import pandas as pd

class ChromosomeGVCF:

    def __init__(self, db, chrom, reference, debug=False):
        self.db = db
        self.chrom = chrom
        self.chrom_length = None
        self.chrom_target_end = None
        self.sample_id = ''
        self.formatted_sample_id = ''
        self.sample_cohort = ''
        self.missing_regions = []
        self.target_regions = []
        self.variant_positions = []
        self.column_headers = ['Identifier', '#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT']
        self.chrom_gvcf = None
        self.reference = reference
        self.debug = debug
        self._set_chrom_length(reference)

    def _set_cohort(self):
        res = self.db.positions.find_one({'samples': {'$elemMatch': {'sample': self.sample_id}}}, {'_id': 0, 'cohort_id': 1})
        try:
            self.sample_cohort = res['cohort_id']
            self._set_chrom_target_end(self.reference)
        except TypeError as e:
            print('No variants found')
            sys.exit(1)

    def _set_chrom_length(self, reference):
        res = self.db.gene_reference_index.find_one({'reference': reference}, {'data': {'$elemMatch': {'chrom': self.chrom}}})
        self.chrom_length = res['data'][0]['length']

    def _set_chrom_target_end(self, reference):
        res = self.db.target_regions.aggregate([
            { '$match': { 'cohort_id': self.sample_cohort, 'chrom': self.chrom }},
            { '$group': { '_id': '$chrom', 'end_target_region': { '$max': '$end' } } }
        ])
        for r in res:
            self.chrom_target_end = r['end_target_region']

    def set_chrom(self, chrom):
        self.chrom = chrom

    def set_sample_id(self, sample_id):
        self.sample_id = sample_id
        self.formatted_sample_id = sample_id
        self._set_cohort()

    def set_missing_regions(self, missing_regions):
        self.missing_regions = missing_regions

    def set_target_regions(self, target_regions):
        self.target_regions = target_regions

    def set_variants(self, variants):
        self.variant_positions = variants

    def convert_to_oracle_sample_id_format(self):
        ary = self.sample_id.split('-')
        fname = ''

        for x in ary:
            fname += x
            
        self.formatted_sample_id = '00000' + fname

    def create_chromosome_gvcf(self):
        print('Generating gVCF chr: %s' % self.chrom)
        event_array = self._create_event_array()
        self._create_chrom_gvcf(event_array)
        print('Chrom %s gVCF created' % self.chrom)

    def _create_event_array(self):

        ############################
        # add 1 for 1-based indexing
        ############################
        b_missing_regions = [0] * (int(self.chrom_length) + 1)
        b_target_regions  = [0] * (int(self.chrom_length) + 1)

        ###############
        # missing array
        ###############
        for missing in self.missing_regions:
            start = missing['start']
            stop  = missing['end']
            b_missing_regions[start:stop+1] = [1] * len(range(start, stop+1))

        ##############
        # target array
        ##############
        for target in self.target_regions:
            start = target['start']
            stop  = target['end']
            b_target_regions[start:stop+1] = [1] * len(range(start, stop+1))

        event_array = self._collapse_missing_and_target_arrays(b_missing_regions, b_target_regions)

        #################################################
        # create variant objects to append to event_array
        #################################################
        var_obj_array = self._create_variant_object_array()

        ##############
        # set variants
        ##############
        firstPass = True
        for var_obj in var_obj_array:
            var_pos = int(var_obj['POS'])

            if event_array[var_pos-1] != 'R' and event_array[var_pos-1] != 'MS':
                event_offset = 0
                while True:
                    if event_array[var_pos + event_offset - 1] == 'R' or event_array[var_pos + event_offset - 1] == 'MS':
                        event_array[var_pos + event_offset - 1] = var_obj
                        break

                    event_offset += 1
            else:
                event_array[var_pos-1] = var_obj                

        if self.debug:
            pthdir = os.path.abspath(os.path.dirname('ChromosomeGVCF.py'))
            with open(pthdir + '/event_stack_chr_' + self.chrom + '.txt', 'w') as f:
                for e in event_array:
                    f.write('%s\n' % e)

        return event_array

    def _collapse_missing_and_target_arrays(self, b_missing_regions, b_target_regions):
        event_array = []
        events = {'missing': 'MS', 'reference': 'R'}

        for i in range(1, len(b_missing_regions)):
            if b_target_regions[i] == 1 and b_missing_regions[i] == 0:
                event_array.append(events['reference'])

            else:
                event_array.append(events['missing'])

        return event_array

    def _create_variant_object_array(self):
        var_obj_array = []

        for variant_position in self.variant_positions:
            for sample in variant_position['samples']:
                if sample['sample'] == self.sample_id:
                    for gene in sample['genes']:
                        for protein in gene['proteins']:
                            var_obj_array.append({
                                'Identifier': 'V',
                                '#CHROM': self.chrom,
                                'POS': variant_position['pos'],
                                'ID': '.',
                                'REF': protein['ref'],
                                'ALT': protein['alt'],
                                'QUAL': '.',
                                'FILTER': 'PASS',
                                'INFO': '.',
                                'GENOTYPE': sample['genotype'],
                                'FORMAT': 'GT:AD:DP:GQ:PL',
                                'OTHER': 'otherFormat',
                            })

        return var_obj_array

    def _create_variant_dataframe(self, event_array, events):

        ttheader = ['Identifier', '#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT', 'OTHER']

        vdf = pd.DataFrame([], columns=ttheader)

        for event in event_array:
            if event != events['missing'] and event != events['reference'] and event != None:
                tary = [event['Identifier'],
                        event['#CHROM'],
                        event['POS'],
                        event['ID'],
                        event['REF'],
                        event['ALT'],
                        event['QUAL'],
                        event['FILTER'],
                        event['INFO'],
                        event['FORMAT'],
                        event['OTHER'],
                    ]

                tdf = pd.DataFrame([tary], columns=ttheader)
                vdf = vdf.append([tdf], ignore_index=True)

        return vdf

    def _create_chrom_gvcf(self, event_array):
        events  = { 'missing': 'MS', 'reference': 'R', 'variant': 'V' }
        record  = { 'start': None, 'stop': None}
        block   = { 'type': None, 'lastType': None, 'startPos': 1, 'currentPos': 1 }
        var_ref = { 'var_start_pos': 0, 'max_var_ref_len': 0 }
        output  = []
        firstPass = True

        ############################################
        # shift array right one for 1-based indexing
        ############################################
        event_array.insert(0, None)
        row = 1

        df_variant = self._create_variant_dataframe(event_array, events)

        ##################################################################################
        # get first element to determine which event is first to initialize block variable
        ##################################################################################
        if isinstance(event_array[row], str):
            block['type'] = event_array[row]
            block['lastType'] = event_array[row]
        else:
            block['type'] = event_array[row]['Identifier']
            block['lastType'] = event_array[row]['Identifier']

        while row < len(event_array):

            block['currentPos'] = row

            ########################
            # get current row's type
            ########################
            if isinstance(event_array[row], str):
                block['type'] = event_array[row]
            else:
                block['type'] = event_array[row]['Identifier']

            #######################################################
            # if block type is not the same as last position's type
            # then a new event has been encountered
            # prepare output of block for gVCF
            #######################################################
            if (block['lastType'] != block['type']) or \
               (block['lastType'] == events['variant'] and block['type'] == events['variant']):

                if var_ref['var_start_pos'] != 0:
                    record['start'] = var_ref['var_start_pos'] + var_ref['max_var_ref_len']
                else:
                    record['start'] = block['startPos']

                record['stop'] = block['currentPos'] - 1

                if block['lastType'] == events['missing']:

                    #####################################################################
                    # if missing block is preceded by a variant with reference lenght > 1 
                    # then account for the overlap at the start position
                    #####################################################################
                    if len(output) > 0 and  output[-1][3] != 'X' and output[-1][3] != '.':
                        record['start'] = int(output[-1][1]) + len(output[-1][3])

                    #########################################################
                    # if previous variant block exceeds the following missing
                    # block in length then omit the missing block
                    #########################################################
                    if record['start'] <= record['stop'] and block['lastType'] != events['variant']:
                        output.append(self._create_missing_record(record))
                        var_ref = { 'var_start_pos': 0, 'max_var_ref_len': 0 }

                elif block['lastType'] == events['reference']:

                    if block['type'] == events['variant'] and event_array[row]['POS'] == record['stop']:
                        record['stop'] = event_array[row]['POS'] - 1

                    ##############################################################
                    # only add reference block if reference block stop position is 
                    # greater than reference block start position
                    ##############################################################
                    if record['stop'] >= record['start']:
                        output.append(self._create_non_variant_record(record))
                    
                    var_ref = { 'var_start_pos': 0, 'max_var_ref_len': 0 }

                else: # events['variant']

                    try:
                        ###################################################################################################################
                        # check if next block is non-variant; use [row] instead of [row+1] since variant was captured on previous iteration
                        # if block is variant only calculate longest ref length if variant is at same position as the previous variant
                        ###################################################################################################################
                        if event_array[row] == events['missing'] or event_array[row] == events['reference'] or \
                            (block['lastType'] == events['variant'] and block['type'] == events['variant'] and event_array[row-1]['POS'] == event_array[row]['POS']):
                            
                            ref_len = len(event_array[block['startPos']]['REF'])
                            var_ref['max_var_ref_len'] = ref_len if ref_len > var_ref['max_var_ref_len'] else var_ref['max_var_ref_len']
                            var_ref['var_start_pos'] = event_array[block['startPos']]['POS']

                            ####################################################################
                            # get row offset when more than one variant is at same pos and chrom
                            ####################################################################
                            num_var_at_same_pos = len(df_variant[df_variant['POS'] == block['startPos']]['POS'])

                        else:
                            var_ref = { 'var_start_pos': 0, 'max_var_ref_len': 0 }

                    except TypeError:
                        pass 

                    output.append(self._create_variant_record(event_array[block['startPos']]))
                    
                block['startPos'] = row

            if isinstance(event_array[row], str):
                block['lastType'] = event_array[row]
            else:
                block['lastType'] = event_array[row]['Identifier']

            row += 1

        ##############################################################
        # create block of last event type to the end of the chromosome
        ##############################################################
        record['start'] = block['startPos']
        record['stop']  = self.chrom_length

        if block['lastType'] == events['missing']:
            output.append(self._create_missing_record(record))

        elif block['lastType'] == events['reference']:
            output.append(self._create_non_variant_record(record))

        else: # events['variant']
            output.append(self._create_variant_record(event_array[block['startPos']]))

        header = self.column_headers[1:len(self.column_headers)]
        header.append(self.formatted_sample_id)
        self.chrom_gvcf = pd.DataFrame(output, columns=header)

    def _create_non_variant_record(self, block_range):
        return [
            self.chrom,
            block_range['start'],
            '.','X','.','.','.',
            'END=' + str(block_range['stop']),
            'GT:AD:DP:GQ:PL',
            '0/0:.:.:.:.',
        ]

    def _create_missing_record(self, block_range):

        return [
            self.chrom,
            block_range['start'],
            '.','.','.','.','.',
            'END=' + str(block_range['stop']),
            'GT:AD:DP:GQ:PL',
            './.:.:.:.:.',
        ]
        return x

    def _create_variant_record(self, variant_data):
        return [
            self.chrom,
            variant_data['POS'],
            variant_data['ID'],
            variant_data['REF'],
            variant_data['ALT'],
            variant_data['QUAL'],
            variant_data['FILTER'],
            variant_data['INFO'],
            variant_data['FORMAT'],
            variant_data['GENOTYPE'] + ':.:.:.:.',
        ]

    def _create_target_record(self, pos, startOrStop):
        return [
            startOrStop, 
            self.chrom,
            pos,
            '.','.','.','.','.','.','.',
        ]

    def _get_max_ref_length(self, ref_ary):
        maxlen = 0
        for ref in ref_ary:
            if len(ref) > maxlen:
                maxlen = len(ref)

        return maxlen






