# gVCF Exporter
# Supports parallelized pool of 8 processes

import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))
import json
import argparse
import datetime
import multiprocessing
from share.classes.ConfigReader import ConfigReader
from pymongo import errors as pyErrors
from pymongo import MongoClient
from classes.ChromosomeGVCF import ChromosomeGVCF
from classes.GVCFExportService import GVCFExportService

################################################################################
################################################################################
# initialization

config_reader = ConfigReader(os.path.join(sys.path[0], '../resources/config/config.yaml'))
config_reader.read_config()

GVCF_HEADER_VERSION = '1.0'
POOL_PROCESSES = 8
MONGO_MAX_POOL_SIZE = 250

USER   = config_reader.yaml['mongodb']['user']
PASS   = config_reader.yaml['mongodb']['passwd']
HOST   = config_reader.yaml['mongodb']['host']
PORT   = str(config_reader.yaml['mongodb']['port'])
AUTHDB = config_reader.yaml['mongodb']['authSource']
NSM_DB = config_reader.yaml['mongodb']['nsm_database']
PIERIANDX_DB = config_reader.yaml['mongodb']['pieriandx_database']
AVATAR_DB = config_reader.yaml['mongodb']['avatar_database']

parser = argparse.ArgumentParser(description='gVCF Exporter for NSM Database')
parser.add_argument('-s', '--sample', help='Enter single Sample ID')
parser.add_argument('-i', '--inputfile', help='Path to TXT file containing list of SampleIDs')
parser.add_argument('-r', '--reference', help='Select .fa.fai reference index file from MongoDB')
parser.add_argument('-l', '--listRef', help='List available reference indexes', action='store_true')
parser.add_argument('-o', '--outdir', help='Output directory', default='./')
parser.add_argument('-a', '--avatar', help='Flag if samples are Avatar IDs', action='store_true')
parser.add_argument('-f', '--foundation', help='Flag if samples are Foundation IDs', action='store_true')

args = parser.parse_args()

################################################################################
################################################################################
################################################################################
# functions

def query_missing_regions(db, sample_id, chrom, cohort):
    return db.missing_data_regions.find({'cohort_id': cohort, 'sample': sample_id, 'chrom': chrom}, {'_id', 'start', 'end'}).batch_size(100000)

def query_target_regions(db, chrom, cohort):
    return db.target_regions.find({'cohort_id': cohort, 'chrom': chrom}, {'_id', 'start', 'end'})

def query_variant_positions(db, sample_id, chrom, cohort):
    return db.positions.find({'cohort_id': cohort, 'chrom': chrom, 'samples': {'$elemMatch': {'sample': sample_id}}}).batch_size(10000)

def query_reference_indexes(db):
    return db.gene_reference_index.find({}, {'_id': 0, 'reference': 1})

def query_gvcf_header(db, version):
    return db.gvcf_header.find_one({'version': version}, {'_id': 0, 'header': 1})

def check_if_reference_version_selected(args):
    if not args.reference:
        print('No reference index selected')
        sys.exit(1)

def create_mongo_client():
    client = MongoClient('mongodb://' 
        + USER + ':' 
        + PASS + '@'
        + HOST + ':'
        + PORT + '/?authSource='
        + AUTHDB, maxPoolSize = MONGO_MAX_POOL_SIZE
    )

    return client

def multiproc_gvcf(key, sample_id):
    client = create_mongo_client()
    db = client[NSM_DB]
    chrom = ChromosomeGVCF(db, key.strip('chrom'), args.reference)
    chrom.set_sample_id(sample_id)

    if not args.avatar and not args.foundation:
        chrom.convert_to_oracle_sample_id_format()

    chrom.set_missing_regions(query_missing_regions(db, sample_id, chrom.chrom, chrom.sample_cohort))
    chrom.set_variants(query_variant_positions(db, sample_id, chrom.chrom, chrom.sample_cohort))
    chrom.set_target_regions(query_target_regions(db, chrom.chrom, chrom.sample_cohort))
    chrom.create_chromosome_gvcf()
    client.close()
    output = {}
    output[key] = chrom.chrom_gvcf
    return output

def get_formatted_date():
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.3%f")

################################################################################
################################################################################
################################################################################
# main()

if __name__ == '__main__':

    client = create_mongo_client()
    db     = client[NSM_DB]
    gvcfs_created = []

    if args.sample:
        check_if_reference_version_selected(args)
        sample_ids = [args.sample]

    elif args.inputfile:
        check_if_reference_version_selected(args)
        sample_ids = []
        with open(args.inputfile, 'r') as f:
            raw = f.readlines()
            [sample_ids.append(x.strip()) for x in raw]

    elif args.listRef:
        versions = query_reference_indexes(db)
        print('\nAvailable references:\n')
        for i, version in enumerate(versions):
            print('%s) %s' % (i+1, version['reference']))

        print('\n')
        sys.exit(0)

    else:
        print('No Sample ID input provided')
        sys.exit(1)

    ############################
    # get gVCF header
    ############################
    gvcf_header = query_gvcf_header(db, GVCF_HEADER_VERSION)['header']

    ############################
    # instantiate export service
    ############################
    export = GVCFExportService(db, '1.0', gvcf_header=gvcf_header)

    client2 = MongoClient('mongodb://' +
        USER + ':' +
        PASS + '@' +
        HOST + ':' +
        PORT + '/?authSource=' +
        AUTHDB
    )

    db2 = client2[PIERIANDX_DB]

    for sample_id in sample_ids:

        print('Sample: %s' % sample_id)

        chrom_dict = {
            'chrom1': None,
            'chrom2': None,
            'chrom3': None,
            'chrom4': None,
            'chrom5': None,
            'chrom6': None,
            'chrom7': None,
            'chrom8': None,
            'chrom9': None,
            'chrom10': None,
            'chrom11': None,
            'chrom12': None,
            'chrom13': None,
            'chrom14': None,
            'chrom15': None,
            'chrom16': None,
            'chrom17': None,
            'chrom18': None,
            'chrom19': None,
            'chrom20': None,
            'chrom21': None,
            'chrom22': None,
            'chromX': None,
            'chromY': None,
            'chromMT': None,
        }

        jobs = []
        results = []

        pool = multiprocessing.Pool(POOL_PROCESSES)

        ###########################
        # launch parallel processes
        ###########################
        for key, value in chrom_dict.items():
            job = pool.apply_async(multiproc_gvcf, args=(key, sample_id))
            jobs.append(job)

        #########################
        # collect process outputs
        #########################
        chrom_gvcfs_d = []
        for j in jobs:
            chrom_gvcfs_d.append(j.get())

        #################################################
        # recombine each chromosomes gvcfs into one array
        #################################################
        chrom_gvcfs = []
        for chrom in chrom_dict:
            [ chrom_gvcfs.append(gvcf_d[chrom]) for gvcf_d in chrom_gvcfs_d if chrom in gvcf_d ]
        
        header = list(chrom_gvcfs[0])
        export.set_sample_id(sample_id)

        if not args.avatar and not args.foundation:
            export.convert_to_oracle_sample_id_format()
        
        export.combine_chrom_gvcf(chrom_gvcfs, header)
        export.save_gvcf(args.outdir, header)

        date = get_formatted_date()

        gvcfs_created.append({
                'accessionNumber': sample_id,
                'gvcfCreated': True,
                'dateCreated': date,
            })

    try:
        db2.gvcfsCreated.insert_many(gvcfs_created, ordered=False)
    except pyErrors.BulkWriteError as e:
        pass

    print('Done')


















