# script to check if all chromosomes and bases are 
# present in a gVCF file

import os
import sys
import argparse
import pandas as pd

GVCF_HEADER_ROWS_TO_SKIP = 19
CHROMOSOMES  = [ '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', 
                 '12', '13', '14', '15', '16', '17', '18', '19', '20',
                 '21', '22', 'X', 'Y', 'MT', ] 

parser = argparse.ArgumentParser(description='Script to check if all'\
    ' all chromosome and bases are present in a gVCF file')
parser.add_argument('-d', '--debug', action='store_true', help='Toggle debug mode')
parser.add_argument('-i', '--input', required=True, help='gVCF file')
parser.add_argument('-c', '--chrom_lengths', required=True, 
                    help='Chromosome lengths \t[chrom | length; tab-delimited]')

args = parser.parse_args()

################################################################################
################################################################################
# functions

def read_gvcf(gvcf):
    sampleid = os.path.basename(gvcf).replace('.gvcf', '')
    return pd.read_csv(gvcf, skiprows=GVCF_HEADER_ROWS_TO_SKIP, 
                        header=GVCF_HEADER_ROWS_TO_SKIP+1, sep='\t',
                        dtype={ '#CHROM': str, 'POS': int, 'ID': str, 'REF': str, 'ALT': str, 'QUAL': str, 'FILTER': str, 'INFO': str, 'FORMAT': str, sampleid: str })

def check_chromosomes(gvcf):
    missing_chroms = []

    gvcf_chroms = _get_chromosomes(gvcf)
    passthru = [ missing_chroms.append(chrom) for chrom in CHROMOSOMES if \
                 chrom not in gvcf_chroms ]    

    if len(missing_chroms) > 0:
        print('Missing chromosomes:')
        for i in range(0, len(missing_chroms)):
            if missing_chroms[i] not in ['X', 'Y', 'MT',] and int(missing_chroms[i]) < 10:
                missing_chroms[i] = '0' + missing_chroms[i]

        missing_chroms.sort()
        [ print(chrom.lstrip('0')) for chrom in missing_chroms ]
        sys.exit(11)

def print_missing_message(chromo, position):
    print('Missing base:\n%s:%s' % (chromo, position))
    exit(1) 

def print_start_pos_gt_stop_postion(start):
    print('Start position greater than stop position')
    print('POS: %s' % start)
    exit(1)

def check_bases(df):
    for chrom in CHROMOSOMES:

        FIRST_PASS = True
        G_START = 0
        G_REF   = 0
        G_STOP  = 0

        if chrom == 1 and df.iloc[0,:]['POS'] != 1:
            print_missing_message(chrom, 1)

        tdf = df.loc[df['#CHROM'] == chrom]
        tdf = tdf.reset_index()

        nrows = tdf.shape[0]
        row   = 0
        while row < nrows:
            start = tdf.iloc[row,:]['POS']
            ref = tdf.iloc[row,:]['REF']
            stop = tdf.iloc[row,:]['INFO'].strip('END=')

            if args.debug:
                print('DEBUG| start: %s' % start)
                print('DEBUG| ref  : %s' % ref)
                print('DEBUG| stop : %s' % stop)
                print('DEBUG| G_START: %s' % G_START)
                print('DEBUG| G_REF  : %s' % G_REF)
                print('DEBUG| G_STOP : %s' % G_STOP)
                print('\n\n')

            try: 
                if start > int(stop):
                    print_start_pos_gt_stop_postion(start)
                
            except (TypeError, ValueError):
                pass

            if FIRST_PASS:
                FIRST_PASS = False
                pass

            else:

                # variant block
                if ref != '.' and ref != 'X':

                    max_var_length = 0

                    df_var = tdf[tdf['POS'] == start]
                    for i in range(0, df_var.shape[0]):
                        if len(df_var.iloc[i,:]['REF']) > max_var_length:
                            max_var_length = len(df_var.iloc[i,:]['REF'])

                    stop = start + max_var_length - 1
                    row += df_var.shape[0] - 1

                # ref or missing block
                if G_STOP + 1 != start:
                    print_missing_message(chrom, G_STOP+1)

            G_START = int(start)
            G_REF   = ref 
            G_STOP  = int(stop)

            row += 1

def check_chrom_start_and_stop(df, chr_lengths):

    df_chr_lengths = _read_chrom_lengths(chr_lengths)

    for chrom in CHROMOSOMES:
        tdf = df.loc[df['#CHROM'] == chrom]
        chr_length = df_chr_lengths.loc[df_chr_lengths['#CHROM'] == chrom].iloc[0,:]['LENGTH']

        if int(tdf.iloc[0,:]['POS']) != 1:
            print('Chromosome %s is missing base 1' % chrom)
            sys.exit(1)

        if int(tdf.iloc[tdf.shape[0]-1,:]['INFO'].strip('END=')) != int(chr_length):
            print('Chromosome %s does not end on chromosome length (%s) from %s' % (chrom, chr_length, chr_lengths))
            sys.exit(1)

def _get_chromosomes(gvcf):
    return list(map(str, set( gvcf.iloc[:,0].tolist() )))

def _read_chrom_lengths(chr_lengths):
    return pd.read_csv(chr_lengths, sep='\t', names=['#CHROM', 'LENGTH',])

################################################################################
################################################################################
################################################################################
# main()

df = read_gvcf(args.input)
check_chromosomes(df)
check_bases(df)
check_chrom_start_and_stop(df, args.chrom_lengths)
print('valid gvcf')









