'''
Script to confirm all variants in pieriandx are 
present in:
	1) the negativeStorageModel.positions collection
	2) the output gVCF for each sample
'''

import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '../..'))
import argparse
import pandas as pd
from pymongo import MongoClient
from share.classes.ConfigReader import ConfigReader

parser = argparse.ArgumentParser(description='Confirm variant presence in ' \
	'pieriandx are in negativeStorageModel.positions and gVCF outputs')
parser.add_argument('-s', '--sample', help='Process single accession number')
parser.add_argument('-i', '--infile', help='Process input file of accession numbers')
parser.add_argument('-p', '--panel', required=True, help='PierianDX or FoundationOne sequence panel')
parser.add_argument('-o', '--outfile', required=True, help='Output directory')
parser.add_argument('-g', '--gvcfdir', required=True, help='gVCF directory')
parser.add_argument('-t', '--targetdir', required=True, help='Target beds directory')
parser.add_argument('--avatar', action='store_true', help='Avatar data')
parser.add_argument('--foundation', action='store_true', help='Foundation data')
parser.add_argument('--hist', action='store_true', help='Historic data')
parser.add_argument('--silent', action='store_true', help='Suppress output .csv files')
args = parser.parse_args()

if (args.hist and args.avatar) or (args.hist and args.foundation) or (args.avatar and args.foundation):
	print('Cannot use \'--avatar\', \'--hist\' and \'--foundation\' simultaneously')
	sys.exit(1)

config_reader = ConfigReader(os.path.join(sys.path[0], '../../resources/config/config.yaml'))
config_reader.read_config()

USER   = config_reader.yaml['mongodb']['user']
PASS   = config_reader.yaml['mongodb']['passwd']
HOST   = config_reader.yaml['mongodb']['host']
PORT   = str(config_reader.yaml['mongodb']['port'])
AUTHDB = config_reader.yaml['mongodb']['authSource']
NSM_DB = config_reader.yaml['mongodb']['nsm_database']
PIERIANDX_DB = config_reader.yaml['mongodb']['pieriandx_database'] 
AVATAR_DB = config_reader.yaml['mongodb']['avatar_database'] 
FOUNDATION_DB = config_reader.yaml['mongodb']['foundation_database'] 

client = MongoClient('mongodb://' + USER + ':' + PASS + '@' + HOST + ':' + PORT + '/?authSource=' + AUTHDB)

db_nsm = client[NSM_DB]

if args.avatar:
	db_variants = client[AVATAR_DB]
elif args.foundation:
	db_variants = client[FOUNDATION_DB]
else: # pieriandx
	db_variants = client[PIERIANDX_DB]

samples = []

if args.sample:
	samples = [args.sample]

if args.infile:
	passthrough = [ samples.append(line.strip('\n')) for line in open(args.infile) ]

if args.gvcfdir[-1] == '/':
	args.gvcfdir = '/' + args.gvcfdir.strip('/')

if args.targetdir[-1] == '/':
	args.targetdir = '/' + args.targetdir.strip('/')

print('gVCF dir: %s' % args.gvcfdir)

################################################################################
################################################################################
################################################################################
# functions

def pierian_panel_to_nsm_cohort_id(panel):
	if panel == 'avatar':
		docs = db_nsm.cohorts.find({ 'description': panel }, { '_id': 1 })
		panels = []
		passthru = [ panels.append(doc['_id']) for doc in docs ] 
		return panels
	else:
		doc = db_nsm.cohorts.find_one({ 'description': panel }, { '_id': 1 })
		return [doc['_id']]

def convert_id_to_gvcf_id(sampleid):
	return '00000' + sampleid.replace('-', '')

def get_most_recent_report_id(panel, caseid):

	doc = db_variants.signedOutRecords.find_one({ 'caseId': caseid, 'panel': panel, 'source': 'API' })
	ids = []
	
	[ ids.append(r['reportId']) for r in doc['reports'] ]
	ids.sort(reverse=True)
		
	return ids[0]

def get_positions(cohort_id, sample):
	return db_nsm.positions.aggregate([
				{ '$match': { 'cohort_id': { '$in': nsm_cohort_id }, 'samples.sample': sample }},
				{ 
					'$project': { 
						'chrom': 1,
						'pos': 1,
						'samples': { 
							'$filter': { 
								'input': '$samples',
								'as': 'samples',
								'cond': { '$eq': [ '$$samples.sample', sample ] }
							}
						}
					}
				}
			])

def get_target_bed_dataframe():
	target_bed_name = db_nsm.cohorts.find_one({ 'description': args.panel }, { 'target_regions_file_path': 1 })
	target_bed_name = os.path.basename(target_bed_name['target_regions_file_path'])
	return pd.read_csv(args.targetdir + '/' + target_bed_name, sep='\t', dtype=str,
			header=None, names=['chrom','start','stop'])

def _bedToDict(sampleid):
	if 'avatar' in args.panel or args.foundation:
		target_bed_name = db_nsm.cohorts.find_one({ 'description': args.panel, 'samples': sampleid }, { 'target_regions_file_path': 1 })
	else:
		target_bed_name = db_nsm.cohorts.find_one({ 'description': args.panel }, { 'target_regions_file_path': 1 })
	
	target_bed_name = os.path.basename(target_bed_name['target_regions_file_path'])
	target_path = args.targetdir + '/' + target_bed_name
	print('Target BED: %s' % target_path)
	# load bed file into list of lists
	bed_ary = []
	passthru = [ bed_ary.append(row.strip('\n').split('\t')) for row in open(target_path) ]

	# get unique chromosome list
	chroms = []
	passthru = [ chroms.append(row[0]) for row in bed_ary ]
	chroms = list(set(chroms))

	# create chromosome dictionary
	chrom_dict = {}
	for chrom in chroms:
		chrom_dict[chrom] = None

	# create bases dictionary for each chromosome and save to chrom_dict
	bases_dict = {}
	for key, value in chrom_dict.items():
		
		for row in bed_ary:

			if row[0] == key:

				for i in range(int(row[1])+1, int(row[2])+1):
					bases_dict[i] = True

		chrom_dict[key] = bases_dict
		bases_dict = {}

	return chrom_dict


def check_variants_are_in_target_regions(tvariants, tchrom_dict, addChrPrefix=False):

	if not addChrPrefix:
		chrPrefix = ''
	else:
		chrPrefix = 'chr'

	var_count = 0

	for tvariant in tvariants:

		if ',' in tvariant['alt']:
			addend = len(tvariant['alt'].split(','))
		else:
			addend = 1

		if tvariant['chrom'] == 23 or tvariant['chrom'] == '23':
			s_chrom = 'X'
		elif tvariant['chrom'] == 24 or tvariant['chrom'] == '24':
			s_chrom = 'Y'
		elif tvariant['chrom'] == 'X' or tvariant['chrom'] == 'Y':
			s_chrom = tvariant['chrom']
		else:
			s_chrom = str(int(tvariant['chrom']))

		try:
			if tchrom_dict[chrPrefix + s_chrom][int(tvariant['pos'])]:
				var_count += addend
		except KeyError:
			pass

	return var_count
		
################################################################################
################################################################################
################################################################################
# main()

print('Panel: %s' % args.panel)
headers = ['panel', 'sample', 'pierian/nsm', 'pierian/gvcf', 'nsm/gvcf', 'error',]
gvcf_headers = ['chrom', 'pos', 'id', 'ref', 'alt', 'qual', 'filter', 'info', 'format', 'sample',]
df = pd.DataFrame([], columns=headers)
chrom_dict = _bedToDict(samples[0])
addChrPrefix = True

for sample in samples:

	print('Sample: %s' % sample)
	error_ary = [ args.panel, sample, None, None, None, False ]




	if args.hist: # Historic data
		doc = db_variants.accessionCaseIds.find_one({ 'accessionNumber': sample, 'panel': args.panel, 'source': 'historical' })
		caseid = doc['caseId']
		doc = db_variants.variantCallReports.find_one({ 'caseId': caseid })
		pierian_num_muts = check_variants_are_in_target_regions(doc['mutations'], chrom_dict)

	elif args.avatar:
		docs = db_variants.variants.find({ 'SLID': sample })

		doc_count = 0
		for doc in docs:
			doc_count += 1
		
		if doc_count > 1:
			print('WARNING: SLID %s is present in both SeqCap and IDT cohorts' % sample)
			sys.exit(1)

		pierian_num_muts = check_variants_are_in_target_regions(doc['mutations'], chrom_dict)

	elif args.foundation:
		doc = db_variants.variants.find_one({ 'sampleId': sample, 'kit': args.panel })
		pierian_num_muts = check_variants_are_in_target_regions(doc['mutations'], chrom_dict, addChrPrefix)

	else: # Pierian Non-historic data
		doc = db_variants.accessionCaseIds.find_one({ 'accessionNumber': sample, 'panel': args.panel, 'source': 'API' })
		caseid = doc['caseId']

		reportid = get_most_recent_report_id(args.panel, caseid)

		doc = db_variants.variantCallReports.find_one({ 'caseId': caseid, 'reportId': reportid, 'panel': args.panel, 'source': 'API' })
		pierian_num_muts = check_variants_are_in_target_regions(doc['mutations'], chrom_dict)




	nsm_cohort_id = pierian_panel_to_nsm_cohort_id(args.panel)
	docs = get_positions(nsm_cohort_id, sample)
	
	nsm_num_muts = 0
	for doc in docs:
		for tsample in doc['samples']:
			for gene in tsample['genes']:
				for protein in gene['proteins']:
					nsm_num_muts += 1

	if not args.avatar and not args.foundation:
		gvcf_sample_id = convert_id_to_gvcf_id(sample)
	else:
		gvcf_sample_id = sample
	
	gvcf = pd.read_csv(args.gvcfdir + '/' + gvcf_sample_id + '.gvcf', sep='\t', skiprows=40, names=gvcf_headers, low_memory=False) 
	gvcf_num_muts = len(gvcf['alt'][gvcf['alt'] != '.'])

	if nsm_num_muts != pierian_num_muts:
		print('\tPierian/NSM:\tmismatch')
		print('\t\tPierian:\t%s' % pierian_num_muts)
		print('\t\tNSM:\t\t%s' % nsm_num_muts)
		error_ary[5] = True
		error_ary[2] = str(pierian_num_muts) + '|' + str(nsm_num_muts)

	if pierian_num_muts != gvcf_num_muts:
		print('\tPierian/gVCF:\tmismatch')
		print('\t\tPierian:\t%s' % pierian_num_muts)
		print('\t\tgVCF:\t\t%s' % gvcf_num_muts)
		error_ary[5] = True
		error_ary[3] = str(pierian_num_muts) + '|' + str(gvcf_num_muts)

	if nsm_num_muts != gvcf_num_muts:
		print('\tNSM/gVCF:\tmismatch')
		print('\t\tNSM:\t\t%s' % nsm_num_muts)
		print('\t\tgVCF:\t\t%s' % gvcf_num_muts)
		error_ary[5] = True
		error_ary[4] = str(nsm_num_muts) + '|' + str(gvcf_num_muts)

	tmp = pd.DataFrame([error_ary], columns=headers)
	df  = df.append([tmp], ignore_index=True)

if not args.silent:
	if args.hist:
		df.to_csv(args.outfile + '/confirm_variant_presence_historic.csv', index=False)
	elif args.avatar:
		df.to_csv(args.outfile + '/confirm_variant_presence_avatar.csv', index=False)
	elif args.foundation:
		df.to_csv(args.outfile + '/confirm_variant_presence_' + args.panel + '.csv', index=False)
	else:
		df.to_csv(args.outfile + '/confirm_variant_presence.csv', index=False)




